/*
// A simple C-style interface for polynomials
//
//
//
*/

#include "poly.h"
#include <stdlib.h>

/*
//  static = visibility only in this file
//  inline = expand inline the function when called
//
//  expression ? execute if true : execute if false ;
*/

static
inline
indexType
min( indexType a, indexType b )
{ return a < b ? a : b ; }
// equivalent
//
// if ( a < b ) 
//   return a ;
// else
//   return b ;
//

static
inline
indexType
max( indexType a, indexType b )
{ return a > b ? a : b ; }

/*
// create a polynomial
//
// the command 
//   new T[size]
// allocate memory for size element of type T.
// Return the pointer pointing to the first element
*/
void
PolyCreate( Poly & p, indexType order ) {
  // in pure C allocation is done by malloc
  p.coeffs = new valueType[order+1] ; // allocate memory
  p.order  = order ;
  for ( indexType i = 0 ; i <= p.order ; ++i ) p.coeffs[i] = 0 ;
}

// resize the polynomial
void
PolyResize( Poly & p, indexType neworder ) {
  if ( neworder > p.order ) {
    valueType * bf = p.coeffs ;
    p.coeffs = new valueType[neworder+1] ; // allocate memory
    indexType i = 0 ;
    for ( ; i <= p . order ; ++i ) p . coeffs[i] = bf[i] ;
    for ( ; i <= neworder  ; ++i ) p . coeffs[i] = 0 ;
    delete [] bf ;
  }
  p.order = neworder ;
}

// destroy a polynomial
void
PolyDestroy( Poly & p ) {
  delete [] p.coeffs ; // free memory
  p.order = -1 ;
}

// copy polynomial q = p
void
PolyCopy( Poly const & p, Poly & q ) {
  if ( q.order < p.order ) {
    PolyDestroy(q) ;
    PolyCreate(q,p.order) ;
  }
  q.order = p.order;
  for ( indexType i = 0 ; i <= q.order ; ++i )
    q.coeffs[i] = p.coeffs[i] ;
    //p.coeffs[i] = q.coeffs[i] ; // error not signaled by the compiler
}

// set the i-th coefficient of the polynomial
void
PolySet( Poly & p, valueType v, indexType i ) {
  if ( i < 0 ) {
    printf( "PolySet(p,%f,%d) bad index\n", v, i ) ;
    exit(1) ; // error found, exiting to the program
  }
  if ( i > p.order ) PolyResize( p, i ) ;
  p.coeffs[i] = v ;
}

// set coefficients of the polynomial
void
PolySetValues( Poly & p, valueType v[], indexType nv ) {
  PolyDestroy( p ) ;
  PolyCreate( p, nv-1 ) ;
  for ( indexType i = 0 ; i < nv ; ++i )
    p.coeffs[i] = v[i] ;
}

// get the i-th coefficient of the polynomial
valueType
PolyGet( Poly const & p, indexType i ) {
  if ( i > p . order || i < 0 )
    return 0 ;
  else
    return p . coeffs[i] ;
}

// return the true polynomial order
indexType
PolyGetOrder( Poly const & p ) {
  indexType order = p.order ;
  for ( ; order >= 0 ; --order )
    if ( p.coeffs[order] != 0 )
      break ;
  return order ;
}

// print in a human readable form the polinomial p to the file fd
// +1 + -1 x + -2 * x**2 + 0 * x**3 
void
PolyPrint( Poly & p, FILE * fd ) {
  for ( indexType i = 0 ; i <= p.order ; ++i ) {
    valueType v = p.coeffs[i] ;
    switch (i) {
      case 0:  fprintf( fd, "%g", v ) ; break ;
      case 1:
        // skip zero values, eliminates +-1
        if ( v > 0 ) {
          if ( v == 1 )  fprintf( fd, " + x" ) ;
          else           fprintf( fd, " + %g*x", v ) ;
        } else if ( v < 0 ) {
          if ( v == -1 ) fprintf( fd, " - x" ) ;
          else           fprintf( fd, " - %g*x", -v ) ;
        }
      break ;
      default:
        // skip zero values, eliminates +-1
        if ( v > 0 ) {
          if ( v == 1 )  fprintf( fd, " + x**%d", i ) ;
          else           fprintf( fd, " + %g*x**%d", v,  i ) ;
        } else if ( v < 0 ) {
          if ( v == -1 ) fprintf( fd, " - x**%d", i ) ;
          else           fprintf( fd, " - %g*x**%d", -v, i ) ;
        }
      break ;
    }
  }
  if (  p.order < 0 ) fprintf( fd, "0\n" ) ;
  else                fprintf( fd, "\n" ) ;
}

// multyply a polynomial by a scalar
// p = a * p
void
PolySCAL( Poly & p, valueType a ) {
  for ( indexType i = 0 ; i <= p . order ; ++i )
    p . coeffs[i] *= a ;
}

// add a polynomial times a scalar to the polynomial
// p = p + a * q
void
PolyAXPY( valueType a, Poly const & q, Poly & p ) {
  if ( p . order < q . order ) PolyResize( p, q . order ) ;
  for ( indexType i = 0 ; i <= p . order ; ++i )
    p . coeffs[i] += a * q . coeffs[i] ;
}

// multiply two polynomial
// p(x) = q(x) * r(x)
// p(x) = p0 + p1 * x + p2 * x**2
// q(x) = q0 + q1 * x + q2 * x**2
// p(x)*q(x) = pq0 + pq1 * x + pq2 * x**2 + ... + pq4 * x**4
// pq0 = p0 * q0
// pq1 = p0 * q1 + p1 * q0
// pq2 = p0 * q2 + p1 * q1 + p2 * q0
//
// pqk = sum( i+j = k ) pi * qj
//
void
PolyMUL( Poly const & q, Poly const & r, Poly & p ) {
  PolyResize( p, q.order + r.order ) ;
  for ( indexType i = 0 ; i <= p . order ; ++i ) {
    valueType v = 0 ;
    for ( indexType k = max(0,i-r.order) ; k <= min(i,q.order) ; ++k )
      v += q.coeffs[k] * r.coeffs[i-k] ;
    p.coeffs[i] = v ;
  }
}

// divide two polynomial a/b return the quotient and the rest
// a = p * b + r
void
PolyDIV( Poly const & a, Poly const & b, Poly & p, Poly & r ) {

  // compute true order of a and b
  indexType aOrder = PolyGetOrder(a) ;
  indexType bOrder = PolyGetOrder(b) ;
  
  if ( bOrder < 0 ) {
    printf( "PolyDIV division by zero\n" ) ;
    exit(1) ; // error found, exiting to the program
  }
  
  if ( bOrder > aOrder ) {
    // if order b > order a then
    // p = 0 and r = a 
    PolyDestroy(p) ;
    PolyCreate(p,0) ;
    PolyCopy(a,r) ;
    return ;
  }

  // reserve memory for computation
  PolyDestroy(p) ;
  PolyCreate(p,aOrder-bOrder) ;

  // initialize r = a
  PolyCopy(a,r) ;

  for ( indexType i = aOrder ; i >= bOrder ; --i ) {
    // compute coeff to erase a.coeffs[i]
    valueType cf = r.coeffs[i] / b.coeffs[bOrder] ;
    for ( indexType j = 0 ; j < bOrder ; ++j ) {
      r.coeffs[i+j-bOrder] -= cf * b.coeffs[j] ;
    }
    r.coeffs[i]        = 0 ;
    p.coeffs[i-bOrder] = cf ;
  }

  // drop quasi zero element
  for ( indexType i = 0 ; i < bOrder ; ++i ) {
    valueType & rr = r.coeffs[i] ;
    if ( rr < 1E-10 && rr > -1E-10 ) rr = 0 ;
  }

  // compute true order of p and r
  p.order = PolyGetOrder(p) ;
  r.order = PolyGetOrder(r) ;
}

// compute polynomial derivative
void
PolyDiff( Poly const & p, Poly & q ) {
  PolyDestroy(q) ;
  PolyCreate(q,p.order-1) ;
  for ( indexType i = 0 ; i < p.order ; ++i )
    q.coeffs[i] = p.coeffs[i+1]*(i+1) ;
}

// compute polynomial integral
void
PolyInt( Poly const & p, Poly & q )  {
  PolyDestroy(q) ;
  PolyCreate(q,p.order+1) ;
  q.coeffs[0] = 0 ;
  for ( indexType i = 0 ; i <= p.order ; ++i )
    q.coeffs[i+1] = p.coeffs[i]/(i+1) ;
}

// compute p(v)
// use Horner rule
// p(x) = p0 + p1 * x + p2 * x**2 + p3 * x**3 ;
// p(x) = p0 + x*(p1 + x * (p2 + x * p3)) ;

valueType
PolyEval( Poly const & p, valueType x ) {
  valueType res = p.coeffs[p.order] ;
  for ( indexType i=p.order-1 ; i >= 0 ; --i )
    res = res * x + p.coeffs[i] ;
  return res ;
}

// compute Maximum Common Divisor
void
PolyMCD( Poly const & p,
         Poly const & q,
         indexType  & nPoly,
         Poly         polys[] ) {

  PolyCreate( polys[0], 0 );
  PolyCreate( polys[1], 0 );

  // compute true order of p and r
  indexType pOrder = PolyGetOrder(p);
  indexType qOrder = PolyGetOrder(q);

  if ( pOrder > qOrder ) {
    PolyCopy( p, polys[0] ) ;
    PolyCopy( q, polys[1] ) ;
  } else {
    PolyCopy( q, polys[0] ) ;
    PolyCopy( p, polys[1] ) ;
  }
  Poly buffer ;
  PolyCreate( buffer, 0 );
  for ( nPoly=2 ; polys[nPoly-1] . order > 0 ; ++nPoly ) {
    PolyCreate( polys[nPoly], 0 ) ;
    PolyDIV( polys[nPoly-2], polys[nPoly-1], buffer, polys[nPoly] ) ;
    PolySCAL( polys[nPoly], -1 ) ;
  }
}

// compute Sturm Sequence
void
PolySturm( Poly const & p, indexType & nPoly, Poly polys[] ) {
  Poly p1, A, R ;
  PolyCreate(p1,0) ;
  PolyCreate(A,0) ;
  PolyCreate(R,0) ;

  PolyDiff(p,p1) ; // compute polynomial derivative
  PolyMCD( p, p1, nPoly, polys ) ; // compute MCP between p and p1

  // divide polynomials sequence by MCD and make polynomials monic
  for ( indexType i = 0 ; i < nPoly ; ++i ) {
    // divide two polynomial a/b return the quotient and the rest
    // a = p * b + r
    PolyDIV( polys[i], polys[nPoly-1], A, R ) ;
    PolyCopy( A, polys[i] ) ;
  }
  PolyDestroy(p1) ;
  PolyDestroy(A) ;
  PolyDestroy(R) ;
}

// compute sign changes
indexType
PolySturmChanges( indexType nPoly, Poly const polys[], valueType x ) {
  indexType nChange = 0 ;
  valueType v = PolyEval( polys[0], x) ;
  bool      p = v > 0 ;
  for ( indexType i = 1 ; i < nPoly ; ++i ) {
    valueType newv = PolyEval( polys[i], x) ;
    bool      newp = newv > 0 ;
    if ( newv == 0 ) continue ; // skip
    if ( p != newp ) ++nChange ;
    p = newp ;
    v = newv ;
  }
  return nChange ;
}

// compute sign changes
void
PolyPrintSturmChanges( FILE *     fd,
                       indexType  nPoly,
                       Poly const polys[],
                       valueType  x ) {
  valueType v = PolyEval( polys[0], x) ;
  if ( v > 0 )      fprintf(fd, "[ + " );
  else if ( v < 0 ) fprintf(fd, "[ - " );
  else              fprintf(fd, "[ 0 " );
  for ( indexType i = 0 ; i < nPoly ; ++i ) {
    valueType v = PolyEval( polys[i], x) ;
    if ( v > 0 )      fprintf(fd, ", + " );
    else if ( v < 0 ) fprintf(fd, ", - " );
    else              fprintf(fd, ", 0 " );
  }
  fprintf(fd, "]\n" );
}
