/*
// A simple C-style interface for polynomials.
// It implements some operations on polynomials.
//
*/

#include <stdio.h>

// parametrize type: valueType is the type for floating point values
typedef double valueType ; // in STL value_type

// parametrize type: indexType is the type for integer values
typedef int indexType ;

/*
//  The type struct Poly, contain the needed information 
//  to store a polynomial
//
*/
typedef struct {
  valueType * coeffs ;
  indexType order ;
} Poly ;

/*
//  in the definition we use "const" when argument are input argument.
//  This means that the referring value cannot be changed inside the function.
//  If you try to change the compiler complain about it.
//
//  The "const" keywork act on its left unless const is the first keywork.
//  Some example
//
//  double const * a : a is a pointer pointing to a CONSTANT piece of memory
//                     *a = 1 ; is not allowed
//                     ++a    ; is allowed
//  double * const a : a is a CONSTANT pointer pointing to a NON constant piece of memory
//                     *a = 1 ; is allowed
//                     ++a    ; is not allowed
*/

// create a polynomial
// & means that p is a reference to p.
// reference is a pointer WITHOUT the boring use osf *
//
// For example
//
// void fun( int * pa ) { *pa = 1 ; }
// void fun( int & a  ) { a   = 1 ; }
//
void PolyCreate( Poly & p, indexType order ) ;

// resize the polynomial
void PolyResize( Poly & p, indexType neworder ) ;

// destroy a polynomial
void PolyDestroy( Poly & p ) ;

// copy polinomial q = p
void PolyCopy( Poly const & p, Poly & q ) ;

// set the i-th coefficient of the polynomial
void PolySet( Poly & p, valueType v, indexType i ) ;

// set coefficients of the polynomial
void PolySetValues( Poly & p, valueType v[], indexType nv ) ;

// get the i-th coefficient of the polynomial
valueType PolyGet( Poly const & p, indexType i ) ;

// return the true polynomial order
indexType PolyGetOrder( Poly const & p ) ;

// print in a human readable form the polinomial p to the file fd
void PolyPrint( Poly & p, FILE * fd ) ;

// multyply a polynomial by a scalar
// p = a * p
void PolySCAL( Poly & p, valueType a ) ;

// add a polynomial times a scalar to the polynomial
// p = p + a * q
void PolyAXPY( valueType a, Poly const & q, Poly & p ) ;

// multiply two polynomial
// p = q * r
void PolyMUL( Poly const & q, Poly const & r, Poly & p ) ;

// divide two polynomial a/b return the quotient and the remainder
// a = p * b + r
void PolyDIV( Poly const & a, Poly const &, Poly & p, Poly & r ) ;

// compute polynomial derivative
void PolyDiff( Poly const & p, Poly & q ) ;

// compute polynomial integral
void PolyInt( Poly const & p, Poly & q ) ;

// compute p(x)
valueType PolyEval( Poly const & p, valueType x ) ;

// compute Maximum Common Divisor
void PolyMCD( Poly const & p, Poly const & q, indexType & nPoly, Poly polys[] ) ;

// compute Sturm Sequence
void      PolySturm( Poly const & p, indexType & nPoly, Poly polys[] ) ;
indexType PolySturmChanges( indexType nPoly, Poly const polys[], valueType x ) ;
void      PolyPrintSturmChanges( FILE * fd, indexType nPoly, Poly const polys[], valueType x ) ;
